/***************************************************************************/
/*
 * recv.c --
 *
 * Myrinet ATP-latency test receiver.
 *
 * $Id: recv.c,v 1.4 95/07/24 18:08:27 tfisher Exp Locker: tfisher $
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * For more information: 
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/***************************************************************************/

#include <stdio.h>
#include "atp.h"

/* Externals. */
extern	int	parse_args();
extern	char	*alloc();


int
main(int argc, char *argv[])
{
	int i;
	int unit=0;		/* myrinet device to use */
	int interface=1;	/* myrinet interface to use */
	int size;		/* amount of data per iteration */
	int nloops;		/* number of receive iterations */
	int sessID;		/* unique session descripter */
	char *buff;		/* data buffer */
	char *peer;		/* destination host */

	/* Check for and get command line args. */
	if (parse_args(argc, argv, &peer, &size, &nloops) < 0)
		exit (1);

	/* Initialize. */
	buff = alloc(size);

	/* Open up a session. */
	sessID = atp_open(unit, interface, peer);

	/* Receive the start packet. */
	atp_receive(sessID, buff, 4);

	/* Echo data. */
	for (i=0; i < nloops; i++) {
		atp_receive(sessID, buff, size);
		atp_send(sessID, buff, size);
	}

	/* Print information. */
	atp_print_receive_stats(sessID);

	/* Close-down. */
	atp_close(sessID);
	exit(0);
}
