/***************************************************************************/
/*
 * send.c --
 *
 * Myrinet ATP-latency test generator.
 *
 * $Id: send.c,v 1.4 95/07/24 18:08:28 tfisher Exp Locker: tfisher $
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * For more information: 
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/***************************************************************************/

#include <stdio.h>
#include <string.h>
#include "atp.h"

/* Externals. */
extern	char	*alloc();
extern	void	timersub3();


int
main(int argc, char *argv[])
{
	int i;
	int unit=0;		/* myrinet device to use */
	int interface=1;	/* myrinet interface to use */
	int size;		/* amount of data per iteration */
	int nloops;		/* number of send iterations */
	int sessID;		/* unique session descripter */
	char *peer;		/* destination host */
	char *buff;		/* data buffer */
	double usecs;		/* communication delay in microseconds */
	struct timeval start;	/* first time stamp */
	struct timeval finish;	/* second time stamp */
	struct timeval elapsed;	/* difference between */

	/* Check for and get command line args. */
	if (parse_args(argc, argv, &peer, &size, &nloops) < 0)
		exit (1);

	/* Initialize. */
	buff = alloc(size);

	/* Open up a session. */
	sessID = atp_open(unit, interface, peer);

	/* Start the test. */
	atp_send(sessID, buff, 4);
	gettimeofday(&start, 0);

	/* Send the data. */
	for (i=0; i < nloops; i++) {
		atp_send(sessID, buff, size);
		atp_receive(sessID, buff, size);
	}

	/* Finish the test. */
	gettimeofday(&finish, 0);
	timersub3(&finish, &start, &elapsed);

	/* Get the average latency in micro-seconds. */
	usecs = (double)((elapsed.tv_sec * 1000000) + elapsed.tv_usec) /
	    (nloops * 2);

	/* Print information. */
	printf(">>> send: size=%d, loops=%d, time=%d.%06d, usecs=%.0f\n",
	    size, nloops, (int)elapsed.tv_sec, (int)elapsed.tv_usec, usecs);
	atp_print_send_stats(sessID);

	/* Shutdown the session. */
	atp_close(sessID);
	exit(0);
}
