/***************************************************************************/
/*
 * atp.h --
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * $Id: atp.h,v 1.9 95/06/09 12:03:34 tfisher Exp Locker: tfisher $
 *
 * For more information: 
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/***************************************************************************/

#ifndef _atp_h
#define _atp_h

#include <sys/types.h>
#include <sys/time.h>
#include "atp_api.h"

/* Protocols. */
#define ATP_UDPIP		1
#define ATP_MYRINET		2

/* Constants. */
#define MAX_MSG			(MAX_MTU-PACKET_SIZE)
#define MAX_WINDOW		32

/*
 * Maximum retry values.
 * The wait time for receiving an ACK should be longer (retry value should
 * be higher) than the wait time (retry value) for data packets.
 */
#define MAX_RECV_DAT_RETRY	6
#define MAX_RECV_ACK_RETRY	12

#ifdef Myrinet1
#define MYRINET_CP		"/nfs/atom/myricom/myrinet0.94/lib/lcp.dat"
#endif
#ifdef Myrinet2
#define MYRINET_CP		"/nfs/atom/myricom/myrinet/lib/mcp.dat"
#endif

struct stats {
	int	n_sessions;
	int	n_segments;
	int	n_sends;
	int	n_receives;
	int	n_window_timeouts;
	int	n_ping_sends;
	int	n_ping_acks;
	int	n_ping_timeouts;
	int	n_retransmits;
	int	n_resends;
	int	n_retries;
	int	n_duplicates;
	int	n_stales;
	int	n_singles;
	int	n_zerolen;
	int	n_data_copies;
	int	n_window_acks;
	int	n_negative_acks;
	int	n_duplicate_acks;
	int	n_corrupted_acks;
	int	n_unexpected_recvq;
	int	n_unexpected_sizes;
	int	n_missing_fins;
#define STATS_SIZE sizeof(struct stats)
};

struct session {
	int	unit;
	int	interface;
	int	ID;
	int	proto;
	int	mtu;
	int	flags;
	char	host[64];
	char	host_peer[64];
	u_long	addr;
	u_long	addr_peer;
	int	nbuffs;
	struct	timeval	rtt;			/* round trip time in usecs */
	struct	buffer	*buff[MAX_BUFFS];
	struct	session	*next;
	struct	session	*prev;
	struct	stats	recv_stats;
	struct	stats	send_stats;
#define SESSION_SIZE sizeof(struct session)
};

/* Packet types. */
#define PKT_DAT		0x0001
#define PKT_ACK		0x0002
#define PKT_NACK	0x0004
#define PKT_OOB		0x0008
#define PKT_RDY		0x0010
#define PKT_SYN		0x0020
#define PKT_FIN		0x0040

struct packet {
	short	pkt_flags;	/* packet types */
	short	pkt_size;	/* size of buffer */
	short	pkt_seq;	/* sequence number */
	short	pkt_ack;	/* ack number */
	short	pkt_bun;	/* bundle sequence number */
	short	pkt_win;	/* window size update */
	u_int	pkt_xmit;	/* transmission sequence number */
	u_int	pkt_nacks;	/* bit mask of (un)received pkts */
#define PACKET_SIZE sizeof(struct packet)
};


/* External declarations. */
extern int atp_open();
extern int atp_close();
extern int atp_send();
extern int atp_receive();
extern void atp_print_send_stats();
extern void atp_print_receive_stats();

#endif /* !_atp_h */
