/*****************************************************************************/
/*
 * atp_api.h --
 *
 * $Id: atp_api.h,v 1.3 95/06/01 12:19:04 tfisher Exp Locker: tfisher $
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * The code in this file is derived partially from code (Mt.c) written
 * by David Finucane of Myricom, Inc. which is part of libMt.a a library
 * designed to provide a test interface to the myrinet2 API.
 *
 * For more information: 
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/*****************************************************************************/

#ifndef _AtpApi_h
#define _AtpApi_h

#include <sys/types.h>
#include <sys/time.h>

/* Constants. */
#define MAX_MTU			8192	/* max MTU; must be a multiple of 4 */
#define MAX_BUFFS		30	/* max partitions of global buff */
#define MAX_SESSIONS		2	/* max communication channels */
#define MAX_RECV_QUEUES		4	/* recv queues available from LANai */
#define ACK_BUFF_SIZE		20
#ifdef Myrinet2
#define ACK_QUEUE		0
#define RECEIVE_QUEUE		1
#endif
#define IMMEDIATE_INDEX		99
#define INTERRUPT_INTERVAL	0
#define INTERRUPT_MASK		0

#ifdef Myrinet1
#define LCP_FILE "/nfs/atom/myricom/myrinet0.94/lib/lcp.dat"
#endif
#ifdef Myrinet2
#define MCP_FILE "/nfs/atom/myricom/myrinet/lib/mcp.dat"
#endif

/* Flags. */
#define F_SEND			0x0001
#define F_RECV			0x0004


/*
 * A partition of the system wide piece of shared memory mapped RAM.
 */
struct buffer {
	char	*d;		/* kernel pointer */
	char	*u;		/* user pointer */
	short	size;		/* length of buffer */
	short	used;		/* set if allocated */
	short	owned;		/* set if DMA outstanding */
	short	index;		/* offset within buffer table */
	short	queue;		/* specificly assigned queue */
};

/*
 * Exported routines.
 * ------------------
 */
extern int AtomicInit(
	int unit, 		/* unit to be used */
	int interface, 		/* interface to be initialized */
	char *file,		/* file name of the image to download */
	int mtu);		/* max communication transmission unit */
extern int AtomicOpen(
	int unit,		/* unit we are connected to */
	int interface,		/* interface to be opened */
	int *sessID,		/* will be set to unique session ID */
	int nreq,		/* number of buffers requested */
	struct buffer *buffs[]); /* will be set to an array of DMA buffs */
extern int AtomicClose(
	int unit,		/* unit to be used */
	int interface,		/* interface to be closed */
	int sessID);		/* the unique session descripter */
extern int AtomicReceive(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	struct buffer **buff,	/* dma buffer to receive data into */
	struct timeval *timeout); /* time to spin in u-seconds */
extern int AtomicSend(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_short hostID,		/* destination host ID */
	struct buffer *buff,	/* dma data to send */
	int size);		/* size of data to send */
extern int AtomicSendImmediate(
	int unit,		/* unit to be received from */
	int interface,		/* interface to be used */
	u_short hostID, 	/* destination host ID */
	int queue,		/* queue to communicate over */
	char *buff,		/* data buffer to be sent */
	int size);		/* size of data to send */
extern int AtomicGetSendAck(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	int *index,		/* buffer that was received into */
	int expected);		/* expected ACK number */
extern int AtomicAddReceiveBuff(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_int index,
	u_char *buff,
	u_short size,
	u_short queue);
extern inline int AtomicGetIdFromName(
	char *name);
extern inline int AtomicGetId();


#endif /* !_AtpApi_h */
