/***************************************************************************/
/*
 * atp_api.c --
 *
 * $Id: atp_api.c,v 1.3 95/06/01 12:18:03 tfisher Exp Locker: tfisher $
 *
 * Copyright (c) 1995 by the University of Southern California.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * All rights reserved.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  This software
 * is provided "as is" and without any express or implied warranties,
 * including, without limitation, the implied warranties of merchant-
 * ibility and fitness for a particular purpose.
 *
 * The code in this file is derived partially from code (Mt.c) written
 * by David Finucane of Myricom, Inc. which is part of libMt.a a library 
 * designed to provide a test interface to the myrinet2 API.
 *
 * For more information: 
 * Tom Fisher (tfisher@isi.edu) or Hong Xu (xu@isi.edu)
 */
/***************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <nlist.h>
#include <sys/types.h>
#include <sys/time.h>
#include "config.h"
#include "lanai_dev.h"
#include "myri_api.h"
#include "atp_api.h"

/* Externals. */
extern	void	Print();
extern	void	Error();
extern	int	TimeGet();
extern	int	TimeAdd();
extern	int	TimeCompare();

/* Macros. */
#ifdef Myrinet1
#define RecvQueue(i) ((i % MAX_RECV_QUEUES) + RECV0_Q)
#endif
#ifdef Myrinet2
#define RecvQueue(i) (i)
#endif

/*
 * Exported routines.
 * ------------------
 */
int AtomicInit(
	int unit, 		/* unit to be used */
	int interface, 		/* interface to be initialized */
	char *file,		/* file name of the image to download */
	int mtu);		/* max communication transmission unit */
int AtomicOpen(
	int unit,		/* unit we are connected to */
	int interface,		/* interface to be opened */
	int *sessID,		/* will be set to unique session ID */
	int nreq,		/* number of buffers requested */
	struct buffer *buffs[]); /* will be set to an array of DMA buffs */
int AtomicClose(
	int unit,		/* unit to be used */
	int interface,		/* interface to be closed */
	int sessID);		/* the unique session descripter */
int AtomicReceive(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	struct buffer **buff,	/* dma buffer to receive data into */
	struct timeval *timeout); /* time to spin in u-seconds */
int AtomicSend(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_short hostID,		/* destination host ID */
	struct buffer *buff,	/* dma data to send */
	int size);		/* size of data to send */
int AtomicSendImmediate(
	int unit,		/* unit to be received from */
	int interface,		/* interface to be used */
	u_short hostID, 	/* destination host ID */
	int queue,		/* queue to communicate over */
	char *buff,		/* data buffer to be sent */
	int size);		/* size of data to send */
int AtomicGetSendAck(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	int *index,		/* buffer that was received into */
	int expected);		/* expected ACK number */
int AtomicAddReceiveBuff(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_int index,
	u_char *buff,
	u_short size,
	u_short queue);
inline int AtomicGetIdFromName(
	char *name);
inline int AtomicGetId();


/*
 * Local routines.
 * ---------------
 */
static int AtomicLANaiHandshake(
	int unit, 		/* unit to be used */
	int interface, 		/* interface to be used */
	int commMode,		/* read, write or both */
	int *wasReset);		/* true if LANai just reset */
static inline int AtomicLANaiDoHandshake(
	int unit, 		/* unit to be used */
	int interface);		/* interface to be used */
static void AtomicProcessLANaiToHost(
	int unit,		/* unit to be used */
	int interface); 	/* interface to be used */
static int AtomicAddReceiveBuffs(
	int unit, 		/* unit to be received from */
	int interface, 		/* interface to be received on */
	int reset,		/* if true then do a reset */
	struct buffer buffs[]);
static inline int AtomicSendImmediateData(
	int unit, 		/* unit to be received from */
	int interface, 		/* interface to be received on */
	u_char hostID,
	u_short size,
	u_int index,
	u_short queue,
	char *buff);
static inline int AtomicSendData(
	int unit, 		/* unit to be received from */
	int interface, 		/* interface to be received on */
	u_char hostID,
	u_short size,
	u_short cksumOffset,
	u_short cksumCorrection,
	u_short numGathers,
	Gather *gathers,
	u_short numScatters,
	Scatter *scatters,
	u_int id);
static inline int AtomicSetHostInfo(
	int unit, 		/* unit to be received from */
	int interface, 		/* interface to be received on */
	u_char hardware,
	u_short dmaBurstSize,
	u_short packetType,
	u_short interruptInterval,
	u_short interruptMask);
static inline int AtomicLANaiWasReset(
	int unit, 		/* unit to be received from */
	int interface);		/* interface to be received on */
static inline int AtomicCheckLANaiToHost(
	int unit, 		/* unit to be received from */
	int interface, 		/* interface to be received on */
	u_int *type,
	u_int *arg);
static void AtomicMarkSendBuffs();
static void AtomicClearLANaiAndExit();


/*
 * DEBUG routines.
 * ---------------
 */
static inline void dbugLANaiWasReset(
	char *function,
	int idle_count);
static inline void dbugAddReceiveBuff(
	char *function,
	struct buffer *buff);
static inline void dbugGetSendAck(
	char *function,
	int index);
static inline void dbugReceiveQueueEmpty(
	char *function);


/* 
 * Global variables.
 * -----------------
 */
int	theUnit = 0;
int	theInterface = 0;
int	theNumUnits = 0;
u_char	theMyriID = 0;
int	theMyriLock = 0;
u_short	theDmaBurstSize;
int	theSessions[MAX_SESSIONS];
struct	buffer	theSendBuffs[MAX_BUFFS];
struct	buffer	theRecvBuffs[MAX_BUFFS];
struct	buffer	theAckBuffs[MAX_BUFFS];


/*
 * AtomicInit() --
 * Initializes the Atomic module.  It first figures out how many myrinet
 * interface units are on the SBus, opens the LANai interface board (which 
 * causes the myrinet API code to malloc() and mmap() a chunk of memory 
 * to be shared by both kernel and user address spaces), and determines
 * the current length of this block and the current length of the DMA burst 
 * size.  The global pointers (to this block) UBLOCK and DBLOCK are also
 * set at this time.  Next, the block of shared memory is then partitioned
 * into "MAX_BUFFS" number of global "mtu" sized send and receive buffers.
 * The whole block may become partitioned...  The myrinet interface board 
 * ID (which is the last byte of its IP-address) is retrieved from the API
 * code and saved in the global variable "theMyriID".  Then, if "file"
 * is not NULL, a clearing and reloading of the LANai board's RAM from the 
 * contents of this file is attempted.  And then, regardless of whether
 * the board's RAM was reloaded or not, a reset of the board is done.
 * Lastly, a handshake with the LANai is attempted.  The actions involved
 * in both a handshake and an attempt to do a handshake are described in
 * section that describes the routine "AtomicLANaiHandshake()". This routine
 * then returns a unique session descripter to be used in all subsequent
 * AtomicSend(), AtomicReceive() and AtomicClose() calls.
 */
int
AtomicInit(
	int unit,		/* unit we are connected to */
	int interface,		/* interface we are initializing */
	char *file, 		/* name of LCP (data) file to load */
	int mtu)		/* max transmission unit size to use */
{
	int i;
	int id;
	int size;
 	char *u, *d;
	char *fn = "AtomicInit";

	if (theNumUnits)
		return (0);

 	theUnit = unit;
 	theInterface = interface;
	theNumUnits = open_lanai_copy_block(&size, &theDmaBurstSize);

	if (theNumUnits <= 0)
		Error(1, "%s: No unit found.", fn);
	if (size == 0)
		Error(1, "%s: Unable to allocate space.", fn);
	if (theUnit >= theNumUnits) 
		Error(1, "%s: Bad unit.", fn);
	if (file && *file == 0)
		Error(1, "%s: File undefined.", fn);
 	if (mtu <= 0 || mtu % 4 != 0 || mtu > MAX_MTU)
		Error(1, "%s: Bad MTU %d.", fn, mtu);
	/*if ((theMyriLock = myri_lock(unit)) == 0)
		Error(1, "%s: Couldn't get the myri_lock.", fn);*/

	/* Be able to abort gracefully. */
	signal(SIGINT, AtomicClearLANaiAndExit);

	/*
	 * Partition our space into a chunk of send buffers and
	 * a chunk of receive buffers.
	 */
	d = (char*)DBLOCK[unit];
	u = (char*)UBLOCK[unit];
 
	for (i=0; i < MAX_BUFFS; i++) {
		bzero((char *)&theSendBuffs[i], sizeof(struct buffer));
		bzero((char *)&theRecvBuffs[i], sizeof(struct buffer));
		bzero((char *)&theAckBuffs[i], sizeof(struct buffer));
		theSendBuffs[i].u = u;
		theSendBuffs[i].d = d;
		theSendBuffs[i].index = i;
		theSendBuffs[i].size = mtu;
		theSendBuffs[i].owned = 1; /*ready to use*/
		u += mtu;
		d += mtu;
		theRecvBuffs[i].u = u;
		theRecvBuffs[i].d = d;
		theRecvBuffs[i].index = i;
		theRecvBuffs[i].size = mtu;
		theRecvBuffs[i].owned = 1; /*ready to use*/
		theRecvBuffs[i].queue = RECEIVE_QUEUE;
		u += mtu;
		d += mtu;
		theAckBuffs[i].u = u;
		theAckBuffs[i].d = d;
		theAckBuffs[i].index = i;
		theAckBuffs[i].size = ACK_BUFF_SIZE;
		theAckBuffs[i].owned = 1; /*ready to use*/
		theAckBuffs[i].queue = ACK_QUEUE;
		u += ACK_BUFF_SIZE;
		d += ACK_BUFF_SIZE;
		if (((int)u % 4) ||
		    ((int)d % 4) ||
		    u > (char *)UBLOCK[unit] + size || 
		    d > (char *)DBLOCK[unit] + size) {
			Error(1, "%s: Overshot buff partition.\n", fn);
			break;
		}
	}
 
	/* Get our id. */
	id = AtomicGetId();
	theMyriID = (u_char)(id & 0xff);
	if (theMyriID == 0) 
		Error(1, "%s: No Myrinet ID.", fn);

#ifdef Myrinet1
	/* Reload the LANai. */
	lanai_interrupt_unit(unit, OFF);
 	lanai_reset_unit(unit, ON);	
	if (!clear_load(unit, file))
		AtomicClearLANaiAndExit();
	lanai_reset_unit(unit, OFF);
	sleep(1);
#endif

#ifdef Myrinet2
	/* Reset the interface. */
        myriReset(unit, interface);
	while (!myriWasReset(unit, interface))
		;
#endif

	/* Finally, do a handshake. */
	AtomicLANaiHandshake(unit, interface, F_SEND|F_RECV, NULL);
	return (0);
}


/*
 * AtomicOpen() --
 * This routine opens up a unique session.  The next available session 
 * (if any) is allocated.  A small integer is returned in "sessID" that 
 * represents the unique session.  Also, this routine takes as input a 
 * requested number of buffers.  Each of these buffers is currently 8k bytes.
 * This routine returns the number of buffers safely allocated.
 */
int
AtomicOpen(
	int unit,		/* unit we are connected to */
	int interface,		/* interface we are opening */
	int *sessID,		/* will be set to unique session ID */
	int nreq,		/* number of buffers requested */
	struct buffer *buffs[])	/* will be set to an array of DMA buffs */
{
	int i;
	int nallocated=0;
	char *fn = "AtomicOpen";

	if (theNumUnits == 0)
		Error(1, "%s: LibAtpApi not initialized.", fn);
	if (buffs == 0)
		Error(1, "%s: No buffers provided.", fn);
	if (nreq < 0 || nreq > MAX_BUFFS)
		Error(1, "%s: Bad buffer request.", fn);

	for (i=0; i < MAX_SESSIONS; i++)
		if (theSessions[i] == 0)
			break;
	if (i == MAX_SESSIONS)
		Error(1, "%s: No more sessions.", fn);
	*sessID = i;
	theSessions[i]++;

	for (i=0; i < MAX_BUFFS; i++) {
		if (theSendBuffs[i].used)
			continue;
		if (nallocated == nreq)
			break;
		theSendBuffs[i].used++;
		buffs[nallocated] = &theSendBuffs[i];
		nallocated++;
	}
	return (nallocated);
}


/*
 * AtomicClose() --
 * Close an open session.  If "sessID" is the only session currently open
 * then close down the LANai.
 */
int
AtomicClose(
	int unit,		/* unit to be closed on */
	int interface,		/* interface to be closed on */
	int sessID)		/* the unique session descripter */
{
	int i;
	char *fn = "AtomicClose";

	if (theUnit >= theNumUnits)
		Error(1, "%s: Bad unit.", fn);

	theSessions[sessID] = 0;
	for (i=0; i < MAX_SESSIONS; i++)
		if (theSessions[i])
			break;
	if (i == MAX_SESSIONS) {
		/*clear_load(unit, "0"); XXX */
		/*if (theMyriLock && myri_unlock(unit))
			Print("%s: Can't release myrinet lock: unit %d.",
			    fn, unit);*/
	}
#ifdef Myrinet2
	/* Reset LANai. */
        myriReset(unit, interface);
	while (!myriWasReset(unit, interface))
		;
#endif
	return (0);
}


/*
 * AtomicSend() --
 * Sends a data message of at most MAX_MTU total length onto the myrinet.
 * First a handshake with the LANai is attempted, and if this attempt fails,
 * 0 is returned.  Then, the LANai is queried for any pending send ACKs that
 * may exist so that they can be cleared from the send ACK queue and that
 * the associated outstanding send buffer can be freed (marked as ready to
 * be used again).  Next, the buffer that is expected to be used next is
 * checked, and if it is not marked as "ready" then a 0 is returned.  Next,
 * the header (if any) and data are copied into the send buffer, and one
 * "Gather" structure as well as one "Scatter" structure are initialized
 * with the send buffer and the appropriate queue.  At this point, the data
 * (the "Gather" and "Scatter" structures) are sent into the API code to
 * be sent onto the myrinet towards the host(s) specified by "hostID". If
 * the API send routine fails then a 0 is returned otherwise the send buffer
 * is marked as ready again and a 1 is returned.  Both the header and data 
 * buffers that are passed to this routine must be mutiples of 4 bytes.
 */
int 
AtomicSend(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_short hostID,		/* destination host ID */
	struct buffer *buff, 	/* dma data to send */
	int size)		/* size of data to send */
{
  	Gather gather;
	Scatter scatter;
	char *fn = "AtomicSend";
	static int num_msgs_sent=0;

	if (theNumUnits == 0)
		Error(1, "%s: libAtomic not initialized.", fn);
	if (size > MAX_MTU)
		Error(1, "%s: Bad size param %d.", fn, size);
	if (size % 4)
		Error(1, "%s: Size must be a multiple of 4.", fn, size);
	if (buff != 0 && (int)buff % 4)
		Error(1, "%s: Bad buf param.", fn);

	/* Check our communication with the LANai. */
	if (!AtomicLANaiHandshake(unit, interface, F_SEND, NULL)) 
		return (0);
 
	/* If not ready try later. */
	if (buff->owned == 0) {
		Print("%s: Send buff not ACK'ed yet.\n", fn);
		return (0);
	}

	/* 
	 * Build a trivial scatter and gather pair.
	 */
#ifdef Myrinet1
	scatter.recv_buff_q = RECV0_Q;
	scatter.len_in_bytes = size;
	gather.ptr = buff->d;
	gather.len_in_bytes = size;
#endif
#ifdef Myrinet2
	scatter.receiveQueue = RECEIVE_QUEUE;
	scatter.interface = interface;
	scatter.lengthInBytes = size;
	gather.pointer = buff->d;
	gather.lengthInBytes = size;
#endif

	/*
	 * Send the data -- 
	 * Do not even wait for the initiation of sending. 
	 * This is a totally asynchronous sending of data.
	 */
	if (AtomicSendData(
	    unit,
	    interface,
	    hostID,
	    size,
	    NO_CKSUM, 
	    0, 
	    1, 
	    &gather, 
	    1, 
	    &scatter, 
	    buff->index) <= 0) {
		/* Clear unfinished ACKs. */
		while (AtomicGetSendAck(unit, interface, 0, -1))
			;
		Print("AtomicSendData() failed.\n");
		return (0);
	}
	buff->owned = 0;

	/* Clear unfinished ACKs. */
	while (AtomicGetSendAck(unit, interface, 0, -1))
		;
	num_msgs_sent++;
	return (size);
}


/*
 * AtomicReceive() --
 * Receive a data message from the Myrinet.  A loop is entered that: 1)
 * attempts to do a handshake with the LANai, 2) attempts to read a message 
 * from the myrinet (if a message is pending it is received), and 3) checks 
 * this message's validity.  This loop is exited if either: the handshake 
 * fails or the message acquisition fails (if/when there is no data pending), 
 * or a valid message arrives.  The loop continues upon the receipt of an
 * invalid message.  If the handshake fails then this routine returns 0.  
 * If there is NO data pending, then 0 is also returned.  Only if and when 
 * a valid data message arrives, its length is returned, and a pointer to 
 * the message data is returned in "data".  This memory is then only valid 
 * until the next call to AtomicReceive().
 */
int 
AtomicReceive(
	int unit,		/* unit to be received from */
	int interface,		/* interface to be received over */
	struct buffer **buff,	/* returned as a pointer to shared memory */
	struct timeval *timeout) /* max time to wait in u-seconds */
{
	u_int index;
 	u_int type;
	u_short size;
	u_short checksum;
	u_short queue;
	int rval;
	int idle_count=0;
	int time_active=0;
	struct buffer *bp;
	struct timeval check;
	struct timeval deadline;
	char *fn = "AtomicReceive";
	static int num_msgs_rcvd=0;
 
	if (theNumUnits == 0)
		Error(1, "%s: libAtpApi not initialized.", fn);
	if (buff == 0)
		Error(1, "%s: Data location is NULL.", fn);
	*buff = NULL;

	/*
	 * Establish our give-up time (if any).
	 */
	if (timeout &&
	    (int)timeout != -1 &&
	    TimeCompare(timeout, 0) > 0) {
		time_active++;
		TimeGet(&deadline);
		TimeAdd(&deadline, timeout);
	}

	/* Check our communication with the LANai. */
	if (!AtomicLANaiHandshake(unit, interface, F_RECV, NULL))
		return (-1);

	/*
	 * Wait for first message with an expected id.
	 */
	for (;;) {
		/* See if a message has arrived. */
		rval = AtomicGetReceiveAck(
		    unit, 
		    interface, 
		    &index, 
		    &size, 
		    &checksum, 
		    &type, 
		    &queue);
		if (rval == -1) {
			/* LANai reset occured: reestablish connection. */
			AtomicLANaiHandshake(unit, interface, F_RECV, NULL);
			continue;
		} else if (rval == 0) {
			/* 
			 * Nothing currently in the receive queue: 
			 * If no timeout or time is up then just return.
			 */
			if ((int)timeout == -1) {
				idle_count++;
				continue;
			} else if (timeout == 0 ||
			    TimeCompare(timeout, 0) == 0)
				return (0);
			if (time_active) {
				TimeGet(&check);
				if (TimeCompare(&check, &deadline) > 0)
					return (0);
			}
			idle_count++;
			continue;
		}

		/*
		 * An ACK was received.
		 * Check if the index is corrupt.
		 * Mark the buff as ready to use.
		 * And Give the buff back to the LCP.
		 */
		if (queue == ACK_QUEUE) {
			bp = &theAckBuffs[index];
			if (index != bp->index)
				Error(0, "%s: Corrupted index %d in buffer.",
				    fn, bp->index);
			bp->owned++;
			bp->queue = queue;
  			*buff = bp;
			return (size);
		}

		/*
		 * A normal looking packet was received.
		 * Check if the index is corrupt.  Mark the 
		 * buff as ready to use.  And Give the buff 
		 * back to the LCP.
		 */
		if (queue == RECEIVE_QUEUE) {
			bp = &theRecvBuffs[index];
			if (index != bp->index)
				Error(0, "%s: Corrupted index %d in buffer.",
				    fn, bp->index);
			bp->owned++;
			bp->queue = queue;
  			*buff = bp;
			AtomicProcessLANaiToHost(unit, interface);
			num_msgs_rcvd++;
			return (size);
		}

 		Print("%s: dropped (len=%d; type=%d queue=%d).\n",
		    fn, size, type, queue);
	}
}


static inline int
AtomicSendData(
	int unit,
	int interface,
	u_char hostID,
	u_short size,
	u_short cksumOffset,
	u_short cksumCorrection,
	u_short numGathers,
	Gather *gathers,
	u_short numScatters,
	Scatter *scatters,
	u_int index)
{
	int rval=0;

#ifdef Myrinet1
	rval = myri_cmd_send(
	    unit, hostID, size, cksumOffset, cksumCorrection, 
	    numGathers, gathers, numScatters, scatters, index);
#endif
#ifdef Myrinet2
	rval = myriAddSend(
	    unit, interface, hostID, size, cksumOffset, cksumCorrection, 
	    numGathers, gathers, numScatters, scatters, index);
#endif
	return (rval);
}


/*
 * AtomicSendImmediate() --
 * 	Sends a message (of at most the size of the API send request command 
 * queue which is at most a few hundred bytes) onto the myrinet.  Using this 
 * routine will allow sends to be made with a lower latency than provided 
 * by the general API send (called by AtomicSend()) which is optimized for 
 * bandwidth.  Here (within the API send-immediate call) the message contents 
 * are copied directly into the API send request command queue.
 * 	First a handshake with the LANai is attempted, and if this attempt 
 * fails, 0 is returned.  The LANai is then queried for any pending send ACKs 
 * that may exist so that they can be cleared from the send ACK queue and 
 * that the associated outstanding send buffer can be freed (marked as ready 
 * to be used again).  Next, the buffer that is expected to be used next is
 * checked, and if it is not marked as "ready" then a 0 is returned.  Next,
 * the is data are copied into the send buffer.  And at this point, the data
 * are sent into the API code to be sent onto the myrinet towards the host(s)
 * specified by "hostID". If the API send routine fails then a 0 is returned 
 * otherwise the send buffer is marked as ready again and a 1 is returned.  
 * The data buffer passed to this routine must be a mutiple of 4 bytes.
 */
int 
AtomicSendImmediate(
	int unit,		/* unit to be used */
	int interface,		/* interface to be used */
	u_short hostID, 	/* unique 8-bit destination host ID */
	int queue,		/* specific queue to send on */
	char *buff,		/* data buffer to be sent */
	int size)		/* length of the data */
{
	char *fn = "AtomicSendImmediate";

	if (theUnit >= theNumUnits)
		Error(1, "%s: Bad unit.", fn);
	if (hostID == 0)
		Error(1, "%s: Undefined host.", fn);
	if (size > MAX_MTU || (size % 4))
		Error(1, "%s: Bad length.", fn);
	if (buff != 0 && (int)buff % 4)
		Error(1, "%s: Bad buffer.", fn);

	/* Check our communication with the LANai. */
	if (!AtomicLANaiHandshake(unit, interface, F_SEND, NULL))
		return (0);
 
	/* Clear unfinished ACKs. */
	while (AtomicGetSendAck(unit, interface, 0, -1))
		;

	/*
	 * Send data -- 
	 * wait for the initiation of sending: not for the sending itself.
	 */
	if (AtomicSendImmediateData(
	    unit,
	    interface,
	    hostID,
	    size,
	    IMMEDIATE_INDEX,
	    queue,
	    buff) <= 0)
		return (0);
	return (size);
}

static inline int
AtomicSendImmediateData(
	int unit,
	int interface,
	u_char hostID,
	u_short size,
	u_int index,
	u_short queue,
	char *buff)
{
	int rval=0;

#ifdef Myrinet1
	rval = !myri_cmd_send_imm(
	    unit, hostID, size, index, queue, buff);
#endif
#ifdef Myrinet2
	rval = myriAddSendImmediate(
	    unit, interface, 1, hostID, size, index, queue, (void *)buff, 0);
#endif
	return (rval);
}


/*
 * AtomicGetSendAck() --
 * Wait for the initiation of sending, not for the sending itself.
 * Get a send ACK if one currently exists.  If a send ACK does exist then
 * its associated send buffer is marked as being free or ready to be used
 * again.  This routine returns 1 if a send ACK exists, and 0 if no send
 * ACK exists.
 */
int 
AtomicGetSendAck(
	int unit, 		/* unit we are using */
	int interface, 		/* interface we are using */
	int *index,		/* will be the buff that has been acked */
	int expected) 		/* expected ack number, -1 if none */
{ 
	int rval=0;
	int ackno;
	char *fn = "AtomicGetSendAck";

	if (unit >= theNumUnits)
		Error(1, "%s: Bad unit.", fn);

#ifdef Myrinet1
	rval = myri_cmd_send_ack(unit, &ackno);
#endif
#ifdef Myrinet2
	rval = myriGetSendAck(unit, interface, (u_int *)&ackno);
#endif

	/* If successful, mark the associated buff as ready to use again. */
	if (rval > 0) {
		if (ackno < 0)
		    Error(1, "%s: Bad index.", fn);
		if (expected != -1 &&
		    expected != ackno &&
		    ackno != IMMEDIATE_INDEX)
			Print("%s: ackno=%d index=%d.\n", 
			    fn, expected, ackno);
		/* Ignore the immediate index. */
		if (ackno == IMMEDIATE_INDEX)
			return (1);
		if (expected == ackno || 
		    expected == -1) {
	    		dbugGetSendAck(fn, ackno);
			theSendBuffs[ackno].owned = 1;
			if (index)
			    *index = ackno;
 			return (1);
		}
	} else if (rval == -1) {
	    Print("[%d] %s: LANai reset.\n", getpid(), fn);
	} /*else
	    Print("[%d] %s: myriGetSendAck() failed.\n", getpid(), fn);*/
	return (0);
}


int
AtomicGetReceiveAck(
	int unit,
	int interface,
	u_int *index,
	u_short *len,
	u_short *checksum,
	u_int *type,
	u_short *queue)
{
	int rval;

#ifdef Myrinet1
	rval = myri_cmd_recv_ack(unit, index, len, checksum, type, queue);
#endif
#ifdef Myrinet2
	rval = myriGetReceiveAck(unit, interface, index, len, checksum, type, queue);
#endif
	return (rval);
}


/*
 * AtomicLANaiHandshake() --
 * Listen to the LANai and see if it has been reset and if so attempt to do
 * a handshake with it.  (The handshake involves setting the trigger_val, 
 * listening to the trigger_val, then resetting the trigger_val back to zero, 
 * and then initializing the values of various location variables).  When 
 * the handshake completes and was successfull, the LANai is then set with 
 * specific values for its myrinet ID, DMA burst size, communication packet 
 * type, alarm interval (in seconds), and its interrupt mask.  At this point,
 * regardless of whether the handshake failed or not, the LANai is then 
 * queried to see if any special LANai-to-Host errors may have occured. This 
 * routine then returns; True if the handshake was successful, and False 
 * otherwise.
 */
static int 
AtomicLANaiHandshake(
	int unit, 		/* unit we are connected to */
	int interface, 		/* interface we are using */
	int commMode, 		/* sending or recving or both */
	int *wasReset)		/* (optional) TRUE if unit has been reset */
{
	int i;
	int reset=0;
	int packetType=0;
	char *fn = "AtomicLANaiHandshake";

#ifdef Myrinet1
	packetType = USER_LEVEL_PKT;
#endif
#ifdef Myrinet2
	packetType = UserLevelPacketType;
#endif

	/*
	 * If the LANai was reset: handshake & refresh host information.
	 * In the handshake, wait till complete (if any problem in doing 
	 * so then infinite looping may occur).
	 */
	if ((reset = AtomicLANaiWasReset(unit, interface))) {
		for (i=0; AtomicLANaiDoHandshake(unit, interface) <= 0;)
			i++;
		dbugLANaiWasReset(fn, i);
		AtomicSetHostInfo(
		    unit, interface, theMyriID, theDmaBurstSize, 
		    packetType, INTERRUPT_INTERVAL, INTERRUPT_MASK);

		/* Add the receive buffers. */
		if (commMode & F_RECV) {
			AtomicAddReceiveBuffs(unit, interface, 1, theRecvBuffs);
			AtomicAddReceiveBuffs(unit, interface, 1, theAckBuffs);
		}

		/* Mark send buffs as ready. */
		if (commMode & F_SEND)
			AtomicMarkSendBuffs();

		/* Handle any special LANai-to-Host errors. */
		AtomicProcessLANaiToHost(unit, interface);
 	}

	if (wasReset)
	    *wasReset = reset;

	return (1);
}


static inline int
AtomicSetHostInfo(
	int unit,
	int interface,
	u_char hardware,
	u_short dmaBurstSize,
	u_short packetType,
	u_short interruptInterval,
	u_short interruptMask)
{
	int rval=0;

#ifdef Myrinet1
	rval = myri_cmd_host_info(
	    unit, 
	    hardware, 
	    dmaBurstSize, 
	    packetType, 
	    interruptInterval, 
	    interruptMask);
#endif
#ifdef Myrinet2
	rval = myriSetHostInfo(
	    unit, 
	    interface,
	    hardware, 
	    dmaBurstSize, 
	    packetType, 
	    interruptInterval, 
	    interruptMask);
#endif
	return (rval);
}


static inline int
AtomicLANaiDoHandshake(
	int unit,
	int interface)
{
	int rval=0;

#ifdef Myrinet1
	rval = myri_do_handshake(unit);
#endif
#ifdef Myrinet2
	rval = myriHandshake(unit, interface);
#endif
	return (rval);
}


static inline int
AtomicLANaiWasReset(
	int unit,
	int interface)
{
	int rval=0;

#ifdef Myrinet1
	rval = myri_lanai_was_reset(unit);
#endif
#ifdef Myrinet2
	rval = myriWasReset(unit, interface);
#endif
	return (rval);
}


/*
 * AtomicProcessLANaiToHost() --
 * Check for and appropriately process any messages sent from the LANai 
 * to this host program.
 */
static void
AtomicProcessLANaiToHost(
	int unit,
	int interface)
{
	u_int type, arg;
	char *fn = "AtomicProcessLANaiToHost";

	if (AtomicCheckLANaiToHost(unit, interface, &type, &arg)) {
		switch (type) {
#ifdef Myrinet1
		case RECV_Q_EMPTY:
			AtomicAddReceiveBuffs(unit, interface, 0, theRecvBuffs);
			AtomicAddReceiveBuffs(unit, interface, 0, theAckBuffs);
			lanai_set_ex1_unit(unit, 1);
			break;
		case SEND_ACK_Q_FULL:
 			/* Consume send acks. */
			while (AtomicGetSendAck(unit, interface, 0, -1))
				Print("%s: Got a send ACK.\n", fn);
			break;
		case RECV_ACK_Q_FULL:
 			/*do nothing*/
			break;
#endif
#ifdef Myrinet2
		case ReceiveQueueEmpty:
			dbugReceiveQueueEmpty(fn);
			AtomicAddReceiveBuffs(unit, interface, 0, theRecvBuffs);
			AtomicAddReceiveBuffs(unit, interface, 0, theAckBuffs);
			break;
		case SendAckQueueFull:
 			/* Consume send acks. */
			Error(0, "[%d] %s: SendAckQueueFull.\n", 
			    getpid(), fn);
			while (AtomicGetSendAck(unit, interface, 0, -1))
				Print("%s: Got a send ACK.\n", fn);
			break;
		case ReceiveAckQueueFull:
			Error(0, "[%d] %s: ReceiveAckQueueFull.\n",
			    getpid(), fn);
 			/*do nothing*/
			break;
#endif
		}
	}
}


/*
 * AtomicCheckLANaiToHost() --
 * This routine listens to whatever the LCP has to tell us.
 * This routine then returns -1 on reset, 0 on failure, and 1 on success.
 */
static inline int
AtomicCheckLANaiToHost(
	int unit,
	int interface,
	u_int *type,
	u_int *arg)
{
	int rval=0;

#ifdef Myrinet1
	rval = myri_cmd_check_l2h(unit, type, arg);
#endif
#ifdef Myrinet2
	rval = myriCheckLanaiToHost(unit, interface, type, arg);
#endif
	return (rval);
}


/*
 * AtomicAddReceiveBuffs() --
 * Stuff the LANai with receive buffers. 
 * This routine is called only when we currently have no buffers associated
 * with the LCP (the LCP has no memory pointers to perform DMA transfers into 
 * from its internal receive queues), and only upon reset.  In the latter 
 * case, none of the provided buffers are assumed to be outstanding. Buffers
 * are added into the LCP in a loop and are marked as "ready to use" after
 * they have been added.  If during this loop the LANai becomes completely 
 * "stuffed" (a provided buffer pointer does not point into the currently 
 * allocated shared memory space) then the loop is exited.  After the loop
 * has exited, any remaining unadded buffers are marked as "not usable".
 * This routine then returns the number of buffers added.
 */
static int 
AtomicAddReceiveBuffs(
	int unit, 		/* a LANai board number */
	int interface, 		/* a LANai interface number */
	int reset,		/* if TRUE, called upon reset */
	struct buffer buffs[])	/* array of buffers to add */
{ 
	int index;
	int nbuffs;
	int rval;
	struct buffer *buff;
	char *fn = "AtomicAddReceiveBuffs";

	if (unit >= theNumUnits)
		Error(1, "%s: Bad unit.", fn);

	nbuffs=0;
 	for (index=0; index < MAX_BUFFS; index++) {
		/* 
		 * Ignore outstanding buffers.
		 * Add buffers to the LANai and mark them as ready to use.
		 * Make sure to stop when the lanai has been stuffed.
		 */
		buff = &buffs[index];
		if (buff->owned || reset) {
			if ((rval = AtomicAddReceiveBuff(
			    unit,
			    interface,
			    index,
			    (u_char *)buff->d,
			    (u_short)buff->size,
			    (u_short)buff->queue)) < 0) {
				Error(0, "[%d] %s: Reset: buff=%d, queue=%d.",
				    getpid(), fn, buff->index, buff->queue);
			} else if (rval == 0) {
				Error(0, "[%d] %s: Can't add buff=%d to queue=%d.",
				    getpid(), fn, buff->index, buff->queue);
			} else {
				dbugAddReceiveBuff(fn, buff);
				buff->owned = 0;
				nbuffs++;
			}
		}
 	}

	/*
	 * Mark the rest of the buffers as NOT usable.
	 */
 	for (; index < MAX_BUFFS; index++)
		buffs[index].owned = 0;
	return (nbuffs);
}


/*
 * AtomicAddReceiveBuff() --
 * This routine adds a Receive buffer to a Receive Queue.
 * This routine then returns -1 on reset, 0 on failure, or 1 on success.
 */
int
AtomicAddReceiveBuff(
	int unit,
	int interface,
	u_int index,
	u_char *buff,
	u_short size,
	u_short queue)
{
	int rval=0;
	int loops=0;
	int MaxResets=3;
	char *fn = "AtomicAddReceiveBuff";

	for (;;) {
#ifdef Myrinet1
		rval = myri_cmd_recv_buff(
		    unit, index, buff, size, queue);
#endif
#ifdef Myrinet2
		rval = myriAddReceiveBuffer(
		    unit, interface, index, buff, size, queue);
#endif
		loops++;
		if (rval == -1) {
			Print("[%d] %s: Interface reset.\n", 
			    getpid(), fn);
			if (loops == MaxResets)
				return (0);
			continue;
		} else
		    return (rval);
	}
}


static void 
AtomicMarkSendBuffs()
{
        int i;

        for (i=0; i < MAX_BUFFS; i++)
		if (theSessions[i])
			/* ready to use */
                	theSendBuffs[i].owned = 1;
}


inline int
AtomicGetIdFromName(
	char *name)
{
	int id = -1;
	char *s;

	if (name == 0)
		return (0);
	if (isdigit(*name)) {
		int i=0;
		int cnt=0;
		while (i < strlen(name))
		    if (name[i++] == '.')
			cnt++;
		if (cnt != 3)
			return (0);
		s = (char *)strrchr(name, '.')+1;
		id = atoi(s);
	}

#ifdef Myrinet1
	id = GetMyriIDFromName(name);
#endif
#ifdef Myrinet2
	id = myriAddressFromName(name);
#endif
	return (id);
}


inline int
AtomicGetId()
{
	int id = -1;

#ifdef Myrinet1
	id = (int)myri_get_id();
#endif
#ifdef Myrinet2
	id = myriGetSuggestedAddress();
#endif
	return (id);
}


static void 
AtomicClearLANaiAndExit(void)
{
	int i;
	char *fn = "AtomicClearLANaiAndExit";

	Print("%s: aborting gracefully...", fn);

#ifdef Myrinet2
	/* Reset the interface. */
        myriReset(theUnit, theInterface);
	while (!myriWasReset(theUnit, theInterface))
		;
	/* Wait for any lingering DMAs to finish. */
	for (i=0; i < 12; i++) {
		usleep (160000);
		Print(".");
	}
#endif
#ifdef Myrinet1
	if (theMyriLock && myri_unlock(theUnit))
		Print("%s: Couldn't unlock the myrinet lock file.", fn);
#endif
	Print("done.\n");
	/*clear_load(theUnit, "0"); XXX */
	exit(0);
}


/*
 * DEBUG Routines from this point on:
 * ----------------------------------
 */

static inline void
dbugLANaiWasReset(
	char *function,
	int idle_count)
{
#ifdef DEBUG
	Print("[%d] %s: LANai RESET: iterations=%d to synchronize.\n", 
	    getpid(), function, idle_count);
#endif
}

static inline void
dbugAddReceiveBuff(
	char *function,
	struct buffer *buff)
{
#ifdef DEBUG
	Print("[%d] %s: LANai added receive buff=%d to queue=%d.\n", 
	    getpid(), function, buff->index, buff->queue);
#endif
}

static inline void
dbugGetSendAck(
	char *function,
	int index)
{
#ifdef DEBUGXXX
	Print("[%d] %s: LANai gave back buff=%d.\n", 
	    getpid(), function, index);
#endif
}

static inline void 
dbugReceiveQueueEmpty(
	char *function)
{
#ifdef DEBUG
	Print("[%d] %s: LANai ReceiveQueueEmpty.\n", getpid(), function);
#endif
}
