/*
 * AtomicCopy.c
 * Fast copy.
 * Dr Robert E. Felderman
 * (c) Copyright 1994 Myricom, Inc.
 * All Rights Reserved
 *
 * Copyright (c) 1995 by the University of Southern California
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation in source and binary forms for non-commercial purposes
 * and without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both the copyright notice and
 * this permission notice appear in supporting documentation, and that
 * any documentation, advertising materials, and other materials related
 * to such distribution and use acknowledge that the software was
 * developed by the University of Southern California, Information
 * Sciences Institute.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
 * the suitability of this software for any purpose.  THIS SOFTWARE IS
 * PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: atp_api_copy.c,v 1.2 95/05/16 18:27:43 tfisher Exp Locker: tfisher $
 */

#include <stdio.h>
#include <sys/types.h>

/* Externals. */
extern void bcopy();

/*
 * AtomicCopy() --
 * Will try to do copies using doubles as much as possible.
 * Copies length 2-byte-words from array w1 to array w2.  
 * Overlapping arrays are NOT handled.  This routine is needed 
 * because the VME LANAI device does not implement byte-addressing, 
 * and therefore "bcopy" cannot be used to access shared memory.
 * It returns a pointer to w2[length].  The SBus interface can 
 * handle reading 8b,16b,32b,64b but not writing 8b so we will 
 * use bcopy only for reading from the device.
 */
u_int *
AtomicCopy(
	u_short *w1, 
	u_short *w2, 
	u_int length)
{
	register u_short *rw1 = w1;
	register u_short *rw2 = w2;
	register u_int i;
	register double *d1, *d2;
 	register *i_ptr1, *i_ptr2;
	register len = length, l1 = 0, l2;

	/* Make sure we are correctly aligned. */
	/*if (((unsigned) w1 ^ (unsigned) w2) & 0x7)
		return (bcopy(w1, w2, length));*/

 	/* 
	 * Only the Sbus can handle 64bit reads and writes efficiently.
	 * 64-bit seems to be slow even on SBus, someday we'll have a 
	 * board that correctly handles char->double and we'll use bcopy 
	 * or DMA :-)
	 */
	if (!(((unsigned) w1 ^ (unsigned) w2) & 0x7)) {
		/*
		 * We are aligned on 8 byte boundaries, use doubles for 
		 * copying.  Copy words first to get things aligned.
		 */
		if ((unsigned) w1 % 8)
			for (l1 = i = (8 - ((unsigned) w1 % 8)) >> 1; i; i--)
				*rw2++ = *rw1++;
		d1 = (double *) rw1;
		d2 = (double *) rw2;

		/* Now blast with doubles. */
		for (l2 = i = (len - l1) >> 2; i; i--)
			*d2++ = *d1++;

		rw2 = (unsigned short *) d2;

		/* Clean up any words at the end. */
		if ((len = (len - l1 - (l2 << 2)))) {
			rw1 = (unsigned short *) d1;
			for (i = len; i; i--)
				*rw2++ = *rw1++;
		}
		return ((u_int *)rw2);
	}

	if (!(((unsigned) w1 ^ (unsigned) w2) & 0x3)) {
		/* 
		 * We are aligned on 4 byte boundaries, use ints for 
		 * copying. Copy words first to get things aligned.
		 */
		if ((unsigned) w1 % 4)
			for (l1 = i = (4 - ((unsigned) w1 % 4)) >> 1; i; i--)
				*rw2++ = *rw1++;
		i_ptr1 = (int *) rw1;
		i_ptr2 = (int *) rw2;

		/* Now copy as many ints as we can. */
		for (l2 = i = (len - l1) >> 1; i; i--)
			*i_ptr2++ = *i_ptr1++;

		rw2 = (unsigned short *) i_ptr2;

		/* Clean up any words still remaining. */
		if ((len = (len - l1 - (l2 << 1)))) {
			rw1 = (unsigned short *) i_ptr1;
			for (i = len; i; i--)
				*rw2++ = *rw1++;
		}
		return ((u_int *)rw2);
	}

	/* word alignment only */
	bcopy(w1, w2, length);
	return ((u_int *)w2);
}
